"use strict";
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.diff = diff;
// HACK: Avoid invalid array length
// https://stackoverflow.com/questions/37421749/is-there-an-item-limit-for-large-arrays-in-javascript#:~:text=In%20Chrome%20maximum%20array%20size%20is%20something%20between%201e9%20and%201e10.&text=And%20crashing%20is%20not%20related,with%20a%20normal%20loop%20too
var CHUNK_SIZE = 50000000;
var REMOVED = 1;
var COMMON = 2;
var ADDED = 3;
var chunkIndex = function (ptr) {
    return ~~(ptr / CHUNK_SIZE);
};
function createCommon(A, B, reverse, eq) {
    if (eq === void 0) { eq = function (a, b) { return a === b; }; }
    var common = [];
    if (A.length === 0 || B.length === 0)
        return [];
    for (var i = 0; i < Math.min(A.length, B.length); i += 1) {
        var a = A[reverse ? A.length - i - 1 : i];
        var b = B[reverse ? B.length - i - 1 : i];
        if (eq(a, b)) {
            common.push([a, b]);
        }
        else {
            return common;
        }
    }
    return common;
}
function diff(A, B, eq) {
    var _a;
    if (eq === void 0) { eq = function (a, b) { return a === b; }; }
    function backTrace(A, B, current, swapped) {
        var M = A.length;
        var N = B.length;
        var result = [];
        var a = M - 1;
        var b = N - 1;
        var idChunkIndex = chunkIndex(current.id);
        var j = routes[idChunkIndex][current.id % CHUNK_SIZE];
        var type = types[idChunkIndex][current.id % CHUNK_SIZE];
        while (true) {
            if (!j && !type)
                break;
            var prev = j;
            if (type === REMOVED) {
                result.unshift({ type: (swapped ? 'removed' : 'added'), value: B[b] });
                b -= 1;
            }
            else if (type === ADDED) {
                result.unshift({ type: (swapped ? 'added' : 'removed'), value: A[a] });
                a -= 1;
            }
            else {
                result.unshift({ type: 'common', value: (swapped ? [B[b], A[a]] : [A[a], B[b]]) });
                a -= 1;
                b -= 1;
            }
            var index = chunkIndex(prev);
            j = routes[index][prev % CHUNK_SIZE];
            type = types[index][prev % CHUNK_SIZE];
        }
        return result;
    }
    function createFP(slide, down, k, M) {
        if (slide && slide.y === -1 && down && down.y === -1)
            return { y: 0, id: 0 };
        if ((down && down.y === -1) || k === M || (slide && slide.y) > (down && down.y) + 1) {
            var prev = slide.id;
            ptr++;
            var index = chunkIndex(ptr);
            routes[index][ptr % CHUNK_SIZE] = prev;
            types[index][ptr % CHUNK_SIZE] = ADDED;
            return { y: slide.y, id: ptr };
        }
        else {
            var prev = down.id;
            ptr++;
            var index = chunkIndex(ptr);
            routes[index][ptr % CHUNK_SIZE] = prev;
            types[index][ptr % CHUNK_SIZE] = REMOVED;
            return { y: down.y + 1, id: ptr };
        }
    }
    function snake(k, slide, down, A, B) {
        var M = A.length;
        var N = B.length;
        if (k < -N || M < k)
            return { y: -1 };
        var fp = createFP(slide, down, k, M);
        while (fp.y + k < M && fp.y < N && eq(A[fp.y + k], B[fp.y])) {
            var prev = fp.id;
            ptr++;
            fp.id = ptr;
            fp.y += 1;
            var index = chunkIndex(ptr);
            routes[index][ptr % CHUNK_SIZE] = prev;
            types[index][ptr % CHUNK_SIZE] = COMMON;
        }
        return fp;
    }
    var prefixCommon = createCommon(A, B, undefined, eq);
    var suffixCommon = createCommon(A.slice(prefixCommon.length), B.slice(prefixCommon.length), true, eq).reverse();
    A = suffixCommon.length ? A.slice(prefixCommon.length, -suffixCommon.length) : A.slice(prefixCommon.length);
    B = suffixCommon.length ? B.slice(prefixCommon.length, -suffixCommon.length) : B.slice(prefixCommon.length);
    var swapped = B.length > A.length;
    _a = swapped ? [B, A] : [A, B], A = _a[0], B = _a[1];
    var M = A.length;
    var N = B.length;
    if (!M && !N && !suffixCommon.length && !prefixCommon.length)
        return [];
    if (!N) {
        return __spreadArray(__spreadArray(__spreadArray([], prefixCommon.map(function (_a) {
            var a = _a[0], b = _a[1];
            return ({ type: 'common', value: [a, b] });
        }), true), A.map(function (a) { return ({ type: (swapped ? 'added' : 'removed'), value: a }); }), true), suffixCommon.map(function (_a) {
            var a = _a[0], b = _a[1];
            return ({ type: 'common', value: [a, b] });
        }), true);
    }
    var offset = N;
    var delta = M - N;
    var size = M + N + 1;
    var arrSize = M * N + size + 1;
    var numberOfChunks = Math.ceil(arrSize / CHUNK_SIZE);
    var fp = new Array(size).fill({ y: -1 });
    var routes = __spreadArray([], new Array(numberOfChunks), true).map(function () { return new Array(CHUNK_SIZE); });
    var types = __spreadArray([], new Array(numberOfChunks), true).map(function () { return new Array(CHUNK_SIZE); });
    var ptr = 0;
    var p = -1;
    while (fp[delta + offset].y < N) {
        p = p + 1;
        for (var k = -p; k < delta; ++k) {
            fp[k + offset] = snake(k, fp[k - 1 + offset], fp[k + 1 + offset], A, B);
        }
        for (var k = delta + p; k > delta; --k) {
            fp[k + offset] = snake(k, fp[k - 1 + offset], fp[k + 1 + offset], A, B);
        }
        fp[delta + offset] = snake(delta, fp[delta - 1 + offset], fp[delta + 1 + offset], A, B);
    }
    var pre = prefixCommon.map(function (_a) {
        var a = _a[0], b = _a[1];
        return ({ type: 'common', value: [a, b] });
    });
    var traced = backTrace(A, B, fp[delta + offset], swapped);
    var suf = suffixCommon.map(function (_a) {
        var a = _a[0], b = _a[1];
        return ({ type: 'common', value: [a, b] });
    });
    // cleanup
    routes = null;
    fp = null;
    if ('flat' in Array.prototype) {
        return [pre, traced, suf].flat();
    }
    else {
        return __spreadArray(__spreadArray(__spreadArray([], pre, true), traced, true), suf, true);
    }
}
